require "bundler/vendor/thor/lib/thor/parser/argument"
require "bundler/vendor/thor/lib/thor/parser/arguments"
require "bundler/vendor/thor/lib/thor/parser/option"
require "bundler/vendor/thor/lib/thor/parser/options"
